@@ -92,7 +92,7 @@ static int open_remap_ghost(struct reg_file_info *rfi,
 
 	if (S_ISFIFO(gfe->mode)) {
 		if (mknod(gf->remap.path, gfe->mode, 0)) {
-			pr_perror("Can't create node for ghost file\n");
+			pr_perror("Can't create node for ghost file");
 			goto close_ifd;
 		}
 		ghost_flags = O_RDWR; /* To not block */
@@ -101,12 +101,12 @@ static int open_remap_ghost(struct reg_file_info *rfi,
 
 	gfd = open(gf->remap.path, ghost_flags, gfe->mode);
 	if (gfd < 0) {
-		pr_perror("Can't open ghost file");
+		pr_perror("Can't open ghost file %s", gf->remap.path);
 		goto close_ifd;
 	}
 
 	if (fchown(gfd, gfe->uid, gfe->gid) < 0) {
-		pr_perror("Can't reset user/group on ghost %#x\n", rfe->remap_id);
+		pr_perror("Can't reset user/group on ghost %#x", rfe->remap_id);
 		goto close_all;
 	}
 
@@ -310,6 +310,8 @@ static int dump_ghost_remap(char *path, const struct stat *st, int lfd, u32 id)
 		return -1;
 
 dump_entry:
+	BUG_ON(gf->id & REMAP_GHOST);
+
 	rpe.orig_id = id;
 	rpe.remap_id = gf->id | REMAP_GHOST;
 
@@ -432,56 +434,48 @@ static int check_path_remap(char *rpath, int plen, const struct stat *ost, int l
 
 int dump_one_reg_file(int lfd, u32 id, const struct fd_parms *p)
 {
-	char fd_str[128];
-	char rpath[PATH_MAX + 1] = ".", *path = rpath + 1;
-	int len, rfd;
+	struct fd_link _link, *link;
+	int rfd;
 
 	RegFileEntry rfe = REG_FILE_ENTRY__INIT;
 
-	snprintf(fd_str, sizeof(fd_str), "/proc/self/fd/%d", lfd);
-	len = readlink(fd_str, path, sizeof(rpath) - 2);
-	if (len < 0) {
-		pr_perror("Can't readlink %s", fd_str);
-		return len;
-	}
+	if (!p->link) {
+		if (fill_fdlink(lfd, p, &_link))
+			return -1;
+		link = &_link;
+	} else
+		link = p->link;
 
-	path[len] = '\0';
 	pr_info("Dumping path for %d fd via self %d [%s]\n",
-			p->fd, lfd, path);
+			p->fd, lfd, &link->name[1]);
 
 	/*
 	 * The regular path we can handle should start with slash.
 	 */
-	if (path[0] != '/') {
-		pr_err("The path [%s] is not supported\n", path);
+	if (link->name[1] != '/') {
+		pr_err("The path [%s] is not supported\n", &link->name[1]);
 		return -1;
 	}
 
-	if (check_path_remap(rpath, len, &p->stat, lfd, id))
+	if (check_path_remap(link->name, link->len, &p->stat, lfd, id))
 		return -1;
 
 	rfe.id		= id;
 	rfe.flags	= p->flags;
 	rfe.pos		= p->pos;
 	rfe.fown	= (FownEntry *)&p->fown;
-	rfe.name	= path;
+	rfe.name	= &link->name[1];
 
 	rfd = fdset_fd(glob_fdset, CR_FD_REG_FILES);
 
 	return pb_write_one(rfd, &rfe, PB_REG_FILES);
 }
 
-static const struct fdtype_ops regfile_ops = {
+const struct fdtype_ops regfile_dump_ops = {
 	.type		= FD_TYPES__REG,
 	.dump		= dump_one_reg_file,
 };
 
-int dump_reg_file(struct fd_parms *p, int lfd,
-			     const int fdinfo)
-{
-	return do_dump_gen_file(p, lfd, &regfile_ops, fdinfo);
-}
-
 static int open_path(struct file_desc *d,
 		int(*open_cb)(struct reg_file_info *, void *), void *arg)
 {
@@ -493,7 +487,7 @@ static int open_path(struct file_desc *d,
 	if (rfi->remap) {
 		mutex_lock(ghost_file_mutex);
 		if (link(rfi->remap->path, rfi->path) < 0) {
-			pr_perror("Can't link %s -> %s\n",
+			pr_perror("Can't link %s -> %s",
 					rfi->remap->path, rfi->path);
 			return -1;
 		}
