@@ -32,6 +32,7 @@
 #include "eventpoll.h"
 #include "fsnotify.h"
 #include "signalfd.h"
+#include "namespaces.h"
 
 #include "parasite.h"
 #include "parasite-syscall.h"
@@ -135,13 +136,29 @@ int do_dump_gen_file(struct fd_parms *p, int lfd,
 	return pb_write_one(fdinfo, &e, PB_FDINFO);
 }
 
+int fill_fdlink(int lfd, const struct fd_parms *p, struct fd_link *link)
+{
+	size_t len;
+
+	link->name[0] = '.';
+
+	len = read_fd_link(lfd, &link->name[1], sizeof(link->name) - 1);
+	if (len < 0) {
+		pr_err("Can't read link for pid %d fd %d\n", p->pid, p->fd);
+		return -1;
+	}
+
+	link->len = len + 1;
+	return 0;
+}
+
 static int fill_fd_params(struct parasite_ctl *ctl, int fd, int lfd,
 				struct fd_opts *opts, struct fd_parms *p)
 {
 	int ret;
 
 	if (fstat(lfd, &p->stat) < 0) {
-		pr_perror("Can't stat fd %d\n", lfd);
+		pr_perror("Can't stat fd %d", lfd);
 		return -1;
 	}
 
@@ -164,7 +181,7 @@ static int fill_fd_params(struct parasite_ctl *ctl, int fd, int lfd,
 
 	ret = fcntl(lfd, F_GETSIG, 0);
 	if (ret < 0) {
-		pr_perror("Can't get owner signum on %d\n", lfd);
+		pr_perror("Can't get owner signum on %d", lfd);
 		return -1;
 	}
 	p->fown.signum = ret;
@@ -190,17 +207,22 @@ static int dump_unsupp_fd(const struct fd_parms *p)
 static int dump_chrdev(struct fd_parms *p, int lfd, const int fdinfo)
 {
 	int maj = major(p->stat.st_rdev);
+	const struct fdtype_ops *ops;
 
 	switch (maj) {
 	case MEM_MAJOR:
-		return dump_reg_file(p, lfd, fdinfo);
+		ops = &regfile_dump_ops;
+		break;
 	case TTYAUX_MAJOR:
 	case UNIX98_PTY_MASTER_MAJOR ... (UNIX98_PTY_MASTER_MAJOR + UNIX98_PTY_MAJOR_COUNT - 1):
 	case UNIX98_PTY_SLAVE_MAJOR:
-		return dump_tty(p, lfd, fdinfo);
+		ops = &tty_dump_ops;
+		break;
+	default:
+		return dump_unsupp_fd(p);
 	}
 
-	return dump_unsupp_fd(p);
+	return do_dump_gen_file(p, lfd, ops, fdinfo);
 }
 
 #ifndef PIPEFS_MAGIC
@@ -210,8 +232,9 @@ static int dump_chrdev(struct fd_parms *p, int lfd, const int fdinfo)
 static int dump_one_file(struct parasite_ctl *ctl, int fd, int lfd, struct fd_opts *opts,
 		       const int fdinfo)
 {
-	struct fd_parms p;
+	struct fd_parms p = FD_PARMS_INIT;
 	struct statfs statfs;
+	const struct fdtype_ops *ops;
 
 	if (fill_fd_params(ctl, fd, lfd, opts, &p) < 0) {
 		pr_perror("Can't get stat on %d", fd);
@@ -225,33 +248,50 @@ static int dump_one_file(struct parasite_ctl *ctl, int fd, int lfd, struct fd_op
 		return dump_chrdev(&p, lfd, fdinfo);
 
 	if (fstatfs(lfd, &statfs)) {
-		pr_perror("Can't obtain statfs on fd %d\n", fd);
+		pr_perror("Can't obtain statfs on fd %d", fd);
 		return -1;
 	}
 
 	if (is_anon_inode(&statfs)) {
 		if (is_eventfd_link(lfd))
-			return dump_eventfd(&p, lfd, fdinfo);
+			ops = &eventfd_dump_ops;
 		else if (is_eventpoll_link(lfd))
-			return dump_eventpoll(&p, lfd, fdinfo);
+			ops = &eventpoll_dump_ops;
 		else if (is_inotify_link(lfd))
-			return dump_inotify(&p, lfd, fdinfo);
+			ops = &inotify_dump_ops;
 		else if (is_fanotify_link(lfd))
-			return dump_fanotify(&p, lfd, fdinfo);
+			ops = &fanotify_dump_ops;
 		else if (is_signalfd_link(lfd))
-			return dump_signalfd(&p, lfd, fdinfo);
+			ops = &signalfd_dump_ops;
 		else
 			return dump_unsupp_fd(&p);
+
+		return do_dump_gen_file(&p, lfd, ops, fdinfo);
 	}
 
-	if (S_ISREG(p.stat.st_mode) || S_ISDIR(p.stat.st_mode))
-		return dump_reg_file(&p, lfd, fdinfo);
+	if (S_ISREG(p.stat.st_mode) || S_ISDIR(p.stat.st_mode)) {
+		struct fd_link link;
+
+		if (fill_fdlink(lfd, &p, &link))
+			return -1;
+
+		p.link = &link;
+		if (link.name[1] == '/')
+			return do_dump_gen_file(&p, lfd, &regfile_dump_ops, fdinfo);
+
+		if (check_ns_proc(&link))
+			return do_dump_gen_file(&p, lfd, &nsfile_dump_ops, fdinfo);
+
+		return dump_unsupp_fd(&p);
+	}
 
 	if (S_ISFIFO(p.stat.st_mode)) {
 		if (statfs.f_type == PIPEFS_MAGIC)
-			return dump_pipe(&p, lfd, fdinfo);
+			ops = &pipe_dump_ops;
 		else
-			return dump_fifo(&p, lfd, fdinfo);
+			ops = &fifo_dump_ops;
+
+		return do_dump_gen_file(&p, lfd, ops, fdinfo);
 	}
 
 	return dump_unsupp_fd(&p);
