@@ -13,6 +13,7 @@
 #include "mount.h"
 #include "net.h"
 #include "pstree.h"
+#include "stats.h"
 #include "protobuf.h"
 #include "protobuf/inventory.pb-c.h"
 #include "protobuf/pagemap.pb-c.h"
@@ -73,8 +74,10 @@ int write_img_inventory(void)
 
 	crt.state = TASK_ALIVE;
 	crt.pid.real = getpid();
-	if (get_task_ids(&crt))
+	if (get_task_ids(&crt)){
+		close(fd);
 		return -1;
+	}
 
 	he.root_ids = crt.ids;
 
@@ -92,82 +95,12 @@ void kill_inventory(void)
 			fdset_template[CR_FD_INVENTORY].fmt, 0);
 }
 
-static void show_inventory(int fd, struct cr_options *o)
+void show_inventory(int fd)
 {
 	pb_show_vertical(fd, PB_INVENTORY);
 }
 
-/*
- * The cr fd set is the set of files where the information
- * about dumped processes is stored. Each file carries some
- * small portion of info about the whole picture, see below
- * for more details.
- */
-
-#define FD_ENTRY(_name, _fmt, _show)		\
-	[CR_FD_##_name] = {			\
-		.fmt	= _fmt ".img",		\
-		.magic	= _name##_MAGIC,	\
-		.show	= _show,		\
-	}
-
-static void show_raw_image(int fd, struct cr_options *opt) {};
-
-struct cr_fd_desc_tmpl fdset_template[CR_FD_MAX] = {
-	FD_ENTRY(INVENTORY,	"inventory",	 show_inventory),
-	FD_ENTRY(FDINFO,	"fdinfo-%d",	 show_files),
-	FD_ENTRY(PAGEMAP,	"pagemap-%ld",	 show_pagemap),
-	FD_ENTRY(SHMEM_PAGEMAP,	"pagemap-shmem-%ld", show_pagemap),
-	FD_ENTRY(REG_FILES,	"reg-files",	 show_reg_files),
-	FD_ENTRY(EVENTFD,	"eventfd",	 show_eventfds),
-	FD_ENTRY(EVENTPOLL,	"eventpoll",	 show_eventpoll),
-	FD_ENTRY(EVENTPOLL_TFD,	"eventpoll-tfd", show_eventpoll_tfd),
-	FD_ENTRY(SIGNALFD,	"signalfd",	 show_signalfd),
-	FD_ENTRY(INOTIFY,	"inotify",	 show_inotify),
-	FD_ENTRY(INOTIFY_WD,	"inotify-wd",	 show_inotify_wd),
-	FD_ENTRY(FANOTIFY,	"fanotify",	 show_fanotify),
-	FD_ENTRY(FANOTIFY_MARK,	"fanotify-mark", show_fanotify_mark),
-	FD_ENTRY(CORE,		"core-%d",	 show_core),
-	FD_ENTRY(IDS,		"ids-%d",	 show_ids),
-	FD_ENTRY(MM,		"mm-%d",	 show_mm),
-	FD_ENTRY(VMAS,		"vmas-%d",	 show_vmas),
-	FD_ENTRY(PIPES,		"pipes",	 show_pipes),
-	FD_ENTRY(PIPES_DATA,	"pipes-data",	 show_pipes_data),
-	FD_ENTRY(FIFO,		"fifo",		 show_fifo),
-	FD_ENTRY(FIFO_DATA,	"fifo-data",	 show_fifo_data),
-	FD_ENTRY(PSTREE,	"pstree",	 show_pstree),
-	FD_ENTRY(SIGACT,	"sigacts-%d",	 show_sigacts),
-	FD_ENTRY(UNIXSK,	"unixsk",	 show_unixsk),
-	FD_ENTRY(INETSK,	"inetsk",	 show_inetsk),
-	FD_ENTRY(PACKETSK,	"packetsk",	 show_packetsk),
-	FD_ENTRY(NETLINKSK,	"netlinksk",	 show_netlinksk),
-	FD_ENTRY(SK_QUEUES,	"sk-queues",	 show_sk_queues),
-	FD_ENTRY(ITIMERS,	"itimers-%d",	 show_itimers),
-	FD_ENTRY(CREDS,		"creds-%d",	 show_creds),
-	FD_ENTRY(UTSNS,		"utsns-%d",	 show_utsns),
-	FD_ENTRY(IPCNS_VAR,	"ipcns-var-%d",	 show_ipc_var),
-	FD_ENTRY(IPCNS_SHM,	"ipcns-shm-%d",	 show_ipc_shm),
-	FD_ENTRY(IPCNS_MSG,	"ipcns-msg-%d",	 show_ipc_msg),
-	FD_ENTRY(IPCNS_SEM,	"ipcns-sem-%d",	 show_ipc_sem),
-	FD_ENTRY(FS,		"fs-%d",	 show_fs),
-	FD_ENTRY(REMAP_FPATH,	"remap-fpath",	 show_remap_files),
-	FD_ENTRY(GHOST_FILE,	"ghost-file-%x", show_ghost_file),
-	FD_ENTRY(TCP_STREAM,	"tcp-stream-%x", show_tcp_stream),
-	FD_ENTRY(MOUNTPOINTS,	"mountpoints-%d", show_mountpoints),
-	FD_ENTRY(NETDEV,	"netdev-%d",	 show_netdevices),
-	FD_ENTRY(IFADDR,	"ifaddr-%d",	 show_raw_image),
-	FD_ENTRY(ROUTE,		"route-%d",	 show_raw_image),
-	FD_ENTRY(TMPFS,		"tmpfs-%d.tar.gz", show_raw_image),
-	FD_ENTRY(TTY,		"tty",		 show_tty),
-	FD_ENTRY(TTY_INFO,	"tty-info",	 show_tty_info),
-	FD_ENTRY(FILE_LOCKS,	"filelocks-%d",	 show_file_locks),
-	FD_ENTRY(RLIMIT,	"rlimit-%d",	 show_rlimit),
-	FD_ENTRY(PAGES,		"pages-%u",	 NULL),
-	FD_ENTRY(PAGES_OLD,	"pages-%d",	 NULL),
-	FD_ENTRY(SHM_PAGES_OLD, "pages-shmem-%ld", NULL),
-	FD_ENTRY(SIGNAL,	"signal-s-%d",	 show_siginfo), /* shared signals */
-	FD_ENTRY(PSIGNAL,	"signal-p-%d",	 show_siginfo), /* private signals */
-};
+void show_raw_image(int fd) {};
 
 static struct cr_fdset *alloc_cr_fdset(int nr)
 {
@@ -325,8 +258,8 @@ int open_image_dir(void)
 
 	close(fd);
 
-	if (opts.snap_parent) {
-		ret = symlink(opts.snap_parent, CR_PARENT_LINK);
+	if (opts.img_parent) {
+		ret = symlink(opts.img_parent, CR_PARENT_LINK);
 		if (ret < 0) {
 			pr_perror("Can't link parent snapshot.");
 			goto err;
@@ -360,7 +293,7 @@ static unsigned long page_ids = 1;
 void up_page_ids_base(void)
 {
 	/*
-	 * When page server and crtools dump work on
+	 * When page server and criu dump work on
 	 * the same dir, the shmem pagemaps and regular
 	 * pagemaps may have IDs conflicts. Fix this by
 	 * making page server produce page images with
