@@ -4,7 +4,7 @@
 struct fd_parms;
 struct cr_fdset;
 
-extern int dump_fifo(struct fd_parms *p, int lfd, const int fdinfo);
+extern const struct fdtype_ops fifo_dump_ops;
 extern int collect_fifo(void);
 
 #endif /* __CR_FIFO_H__ */
