@@ -30,7 +30,7 @@ extern int collect_reg_files(void);
 
 extern int prepare_shared_reg_files(void);
 
-extern int dump_reg_file(struct fd_parms *p, int lfd, const int fdinfo);
+extern const struct fdtype_ops regfile_dump_ops;
 extern int dump_one_reg_file(int lfd, u32 id, const struct fd_parms *p);
 
 extern struct file_remap *lookup_ghost_remap(u32 dev, u32 ino);
