@@ -12,8 +12,7 @@ extern struct fstype *find_fstype_by_name(char *fst);
 
 struct cr_fdset;
 extern int dump_mnt_ns(int pid, struct cr_fdset *);
-struct cr_options;
-extern void show_mountpoints(int fd, struct cr_options *);
+extern void show_mountpoints(int fd);
 int prepare_mnt_ns(int pid);
 
 extern int pivot_root(const char *new_root, const char *put_old);
