@@ -13,7 +13,6 @@
 
 struct fdinfo_list_entry;
 struct sk_opts_entry;
-struct cr_options;
 struct file_desc;
 struct fd_parms;
 struct cr_fdset;
@@ -43,10 +42,10 @@ extern int resolve_unix_peers(void);
 
 extern int collect_netlink_sockets(void);
 
-extern void show_unixsk(int fd, struct cr_options *o);
-extern void show_inetsk(int fd, struct cr_options *o);
-extern void show_sk_queues(int fd, struct cr_options *o);
-extern void show_netlinksk(int fd, struct cr_options *o);
+extern void show_unixsk(int fd);
+extern void show_inetsk(int fd);
+extern void show_sk_queues(int fd);
+extern void show_netlinksk(int fd);
 
 extern char *skfamily2s(u32 f);
 extern char *sktype2s(u32 t);
@@ -55,10 +54,11 @@ extern char *skstate2s(u32 state);
 
 extern struct socket_desc *lookup_socket(int ino, int family, int proto);
 
-extern int dump_one_inet(struct fd_parms *p, int lfd, const int fdinfo);
-extern int dump_one_inet6(struct fd_parms *p, int lfd, const int fdinfo);
-extern int dump_one_unix(struct fd_parms *p, int lfd, const int fdinfo);
-extern int dump_one_netlink(struct fd_parms *p, int lfd, const int fdinfo);
+extern const struct fdtype_ops unix_dump_ops;
+extern const struct fdtype_ops inet_dump_ops;
+extern const struct fdtype_ops inet6_dump_ops;
+extern const struct fdtype_ops netlink_dump_ops;
+extern const struct fdtype_ops packet_dump_ops;
 
 extern int inet_collect_one(struct nlmsghdr *h, int family, int type);
 extern int unix_receive_one(struct nlmsghdr *h, void *);
