@@ -13,7 +13,7 @@
 #endif
 #define PTS_FMT		"/dev/pts/%d"
 
-extern int dump_tty(struct fd_parms *p, int lfd, const int fdinfo);
+extern const struct fdtype_ops tty_dump_ops;
 extern int dump_verify_tty_sids(void);
 extern int collect_tty(void);
 extern int prepare_shared_tty(void);
@@ -21,7 +21,7 @@ extern int tty_setup_slavery(void);
 
 extern int tty_verify_active_pairs(void);
 
-extern int tty_prep_fds(struct cr_options *opts);
+extern int tty_prep_fds(void);
 extern void tty_fini_fds(void);
 
 #define OPT_SHELL_JOB	"shell-job"
