@@ -11,15 +11,11 @@
 #include "page-xfer.h"
 #include "log.h"
 #include "kerndat.h"
+#include "stats.h"
 
 #include "protobuf.h"
 #include "protobuf/pagemap.pb-c.h"
 
-#define PME_PRESENT	(1ULL << 63)
-#define PME_SWAP	(1ULL << 62)
-#define PME_FILE	(1ULL << 61)
-#define PME_SOFT_DIRTY	(1Ull << 55)
-
 /*
  * On dump we suck in the whole parent pagemap. Then, when observing
  * a page with soft-dirty bit cleared (i.e. -- not modified) we check
@@ -40,14 +36,26 @@ struct mem_snap_ctx {
 #define MEM_SNAP_BATCH	64
 
 static int task_reset_dirty_track(int pid)
+{
+	if (!opts.track_mem)
+		return 0;
+
+	if (!kerndat_has_dirty_track) {
+		pr_err("Kernel doesn't support dirty tracking. "
+				"No snapshot available.\n");
+		return -1;
+	}
+
+	return do_task_reset_dirty_track(pid);
+}
+
+int do_task_reset_dirty_track(int pid)
 {
 	int fd, ret;
 	char cmd[] = "4";
 
-	if (!opts.mem_snapshot)
-		return 0;
-
 	pr_info("Reset %d's dirty tracking\n", pid);
+
 	fd = open_proc_rw(pid, "clear_refs");
 	if (fd < 0)
 		return -1;
@@ -70,15 +78,6 @@ static struct mem_snap_ctx *mem_snap_init(struct parasite_ctl *ctl)
 	int p_fd, pm_fd;
 	PagemapHead *h;
 
-	if (!opts.mem_snapshot)
-		return NULL;
-
-	if (!kerndat_has_dirty_track) {
-		pr_err("Kernel doesn't support dirty tracking. "
-				"No snapshot available.\n");
-		return ERR_PTR(-1);
-	}
-
 	p_fd = get_service_fd(PARENT_FD_OFF);
 	if (p_fd < 0) {
 		pr_debug("Will do full memory dump\n");
@@ -86,8 +85,11 @@ static struct mem_snap_ctx *mem_snap_init(struct parasite_ctl *ctl)
 	}
 
 	pm_fd = open_image_at(p_fd, CR_FD_PAGEMAP, O_RSTR, ctl->pid.virt);
-	if (pm_fd < 0)
+	if (pm_fd < 0) {
+		if (errno == ENOENT)
+			return NULL;
 		return ERR_PTR(pm_fd);
+	}
 
 	ctx = xmalloc(sizeof(*ctx));
 	if (!ctx)
@@ -151,7 +153,7 @@ static void mem_snap_close(struct mem_snap_ctx *ctx)
 	}
 }
 
-unsigned int vmas_pagemap_size(struct vm_area_list *vmas)
+unsigned int dump_pages_args_size(struct vm_area_list *vmas)
 {
 	/*
 	 * In the worst case I need one iovec for half of the
@@ -159,6 +161,7 @@ unsigned int vmas_pagemap_size(struct vm_area_list *vmas)
 	 */
 
 	return sizeof(struct parasite_dump_pages_args) +
+		vmas->nr * sizeof(struct parasite_vma_entry) +
 		(vmas->priv_size + 1) * sizeof(struct iovec) / 2;
 }
 
@@ -212,6 +215,15 @@ static int page_in_parent(unsigned long vaddr, u64 map, struct mem_snap_ctx *sna
 	return 0;
 }
 
+/*
+ * This routine finds out what memory regions to grab from the
+ * dumpee. The iovs generated are then fed into vmsplice to
+ * put the memory into the page-pipe's pipe.
+ *
+ * "Holes" in page-pipe are regions, that should be dumped, but
+ * the memory contents is present in the pagent image set.
+ */
+
 static int generate_iovs(struct vma_area *vma, int pagemap, struct page_pipe *pp, u64 *map,
 		struct mem_snap_ctx *snap)
 {
@@ -252,64 +264,73 @@ static int generate_iovs(struct vma_area *vma, int pagemap, struct page_pipe *pp
 			return -1;
 	}
 
+	cnt_add(CNT_PAGES_SCANNED, nr_to_scan);
+	cnt_add(CNT_PAGES_SKIPPED_PARENT, pages[0]);
+	cnt_add(CNT_PAGES_WRITTEN, pages[1]);
+
 	pr_info("Pagemap generated: %lu pages %lu holes\n", pages[1], pages[0]);
 	return 0;
 }
 
-static int parasite_mprotect_seized(struct parasite_ctl *ctl, struct vm_area_list *vma_area_list, bool unprotect)
+static struct parasite_dump_pages_args *prep_dump_pages_args(struct parasite_ctl *ctl,
+		struct vm_area_list *vma_area_list)
 {
-	struct parasite_mprotect_args *args;
+	struct parasite_dump_pages_args *args;
 	struct parasite_vma_entry *p_vma;
 	struct vma_area *vma;
 
-	args = parasite_args_s(ctl, vmas_pagemap_size(vma_area_list));
+	args = parasite_args_s(ctl, dump_pages_args_size(vma_area_list));
 
-	p_vma = args->vmas;
-	args->nr = 0;
+	p_vma = pargs_vmas(args);
+	args->nr_vmas = 0;
 
 	list_for_each_entry(vma, &vma_area_list->h, list) {
 		if (!privately_dump_vma(vma))
 			continue;
 		if (vma->vma.prot & PROT_READ)
 			continue;
+
 		p_vma->start = vma->vma.start;
 		p_vma->len = vma_area_len(vma);
 		p_vma->prot = vma->vma.prot;
-		if (unprotect)
-			p_vma->prot |= PROT_READ;
-		args->nr++;
+
+		args->nr_vmas++;
 		p_vma++;
 	}
 
-	return parasite_execute(PARASITE_CMD_MPROTECT_VMAS, ctl);
+	return args;
 }
 
 static int __parasite_dump_pages_seized(struct parasite_ctl *ctl,
-		struct vm_area_list *vma_area_list, struct cr_fdset *cr_fdset)
+		struct parasite_dump_pages_args *args,
+		struct vm_area_list *vma_area_list,
+		struct page_pipe **pp_ret)
 {
-	struct parasite_dump_pages_args *args;
 	u64 *map;
 	int pagemap;
 	struct page_pipe *pp;
 	struct page_pipe_buf *ppb;
 	struct vma_area *vma_area;
 	int ret = -1;
-	struct page_xfer xfer;
 	struct mem_snap_ctx *snap;
 
 	pr_info("\n");
 	pr_info("Dumping pages (type: %d pid: %d)\n", CR_FD_PAGES, ctl->pid.real);
 	pr_info("----------------------------------------\n");
 
+	timing_start(TIME_MEMDUMP);
+
 	pr_debug("   Private vmas %lu/%lu pages\n",
 			vma_area_list->longest, vma_area_list->priv_size);
 
+	/*
+	 * Step 0 -- prepare
+	 */
+
 	snap = mem_snap_init(ctl);
 	if (IS_ERR(snap))
 		goto out;
 
-	args = parasite_args_s(ctl, vmas_pagemap_size(vma_area_list));
-
 	map = xmalloc(vma_area_list->longest * sizeof(*map));
 	if (!map)
 		goto out_snap;
@@ -325,10 +346,14 @@ static int __parasite_dump_pages_seized(struct parasite_ctl *ctl,
 	}
 
 	ret = -1;
-	pp = create_page_pipe(vma_area_list->priv_size / 2, args->iovs);
+	pp = create_page_pipe(vma_area_list->priv_size / 2, pargs_iovs(args));
 	if (!pp)
 		goto out_close;
 
+	/*
+	 * Step 1 -- generate the pagemap
+	 */
+
 	list_for_each_entry(vma_area, &vma_area_list->h, list) {
 		if (!privately_dump_vma(vma_area))
 			continue;
@@ -339,35 +364,63 @@ static int __parasite_dump_pages_seized(struct parasite_ctl *ctl,
 	}
 
 	debug_show_page_pipe(pp);
+	
+	/*
+	 * Step 2 -- grab pages into page-pipe
+	 */
 
 	args->off = 0;
 	list_for_each_entry(ppb, &pp->bufs, l) {
+		args->nr_segs = ppb->nr_segs;
+		args->nr_pages = ppb->pages_in;
+		pr_debug("PPB: %d pages %d segs %u pipe %d off\n",
+				args->nr_pages, args->nr_segs, ppb->pipe_size, args->off);
+
+		ret = __parasite_execute_daemon(PARASITE_CMD_DUMPPAGES, ctl, false);
+		if (ret < 0)
+			goto out_pp;
 		ret = parasite_send_fd(ctl, ppb->p[1]);
 		if (ret)
 			goto out_pp;
 
-		args->nr = ppb->nr_segs;
-		args->nr_pages = ppb->pages_in;
-		pr_debug("PPB: %d pages %d segs %u pipe %d off\n",
-				args->nr_pages, args->nr, ppb->pipe_size, args->off);
-
-		ret = parasite_execute(PARASITE_CMD_DUMPPAGES, ctl);
+		ret = __parasite_execute_daemon_wait_ack(PARASITE_CMD_DUMPPAGES, ctl);
 		if (ret < 0)
 			goto out_pp;
 
-		args->off += args->nr;
+		args->off += args->nr_segs;
 	}
 
-	ret = open_page_xfer(&xfer, CR_FD_PAGEMAP, ctl->pid.virt);
-	if (ret < 0)
-		goto out_pp;
+	timing_stop(TIME_MEMDUMP);
 
-	ret = page_xfer_dump_pages(&xfer, pp, 0);
+	/*
+	 * Step 3 -- write pages into image (or delay writing for
+	 *           pre-dump action (see pre_dump_one_task)
+	 */
 
-	xfer.close(&xfer);
-	task_reset_dirty_track(ctl->pid.real);
+	if (pp_ret)
+		*pp_ret = pp;
+	else {
+		struct page_xfer xfer;
+
+		timing_start(TIME_MEMWRITE);
+		ret = open_page_xfer(&xfer, CR_FD_PAGEMAP, ctl->pid.virt);
+		if (ret < 0)
+			goto out_pp;
+
+		ret = page_xfer_dump_pages(&xfer, pp, 0);
+
+		xfer.close(&xfer);
+		timing_stop(TIME_MEMWRITE);
+	}
+
+	/*
+	 * Step 4 -- clean up
+	 */
+
+	ret = task_reset_dirty_track(ctl->pid.real);
 out_pp:
-	destroy_page_pipe(pp);
+	if (ret || !pp_ret)
+		destroy_page_pipe(pp);
 out_close:
 	close(pagemap);
 out_free:
@@ -380,21 +433,34 @@ out:
 }
 
 int parasite_dump_pages_seized(struct parasite_ctl *ctl,
-		struct vm_area_list *vma_area_list, struct cr_fdset *cr_fdset)
+		struct vm_area_list *vma_area_list, struct page_pipe **pp)
 {
 	int ret;
+	struct parasite_dump_pages_args *pargs;
 
-	ret = parasite_mprotect_seized(ctl, vma_area_list, true);
+	pargs = prep_dump_pages_args(ctl, vma_area_list);
+
+	/*
+	 * Add PROT_READ protection for all VMAs we're about to
+	 * dump if they don't have one. Otherwise we'll not be
+	 * able to read the memory contents.
+	 *
+	 * Afterwards -- reprotect memory back.
+	 */
+
+	pargs->add_prot = PROT_READ;
+	ret = parasite_execute_daemon(PARASITE_CMD_MPROTECT_VMAS, ctl);
 	if (ret) {
 		pr_err("Can't dump unprotect vmas with parasite\n");
 		return ret;
 	}
 
-	ret = __parasite_dump_pages_seized(ctl, vma_area_list, cr_fdset);
+	ret = __parasite_dump_pages_seized(ctl, pargs, vma_area_list, pp);
 	if (ret)
 		pr_err("Can't dump page with parasite\n");
 
-	if (parasite_mprotect_seized(ctl, vma_area_list, false)) {
+	pargs->add_prot = 0;
+	if (parasite_execute_daemon(PARASITE_CMD_MPROTECT_VMAS, ctl)) {
 		pr_err("Can't rollback unprotected vmas with parasite\n");
 		ret = -1;
 	}
