@@ -19,6 +19,7 @@
 #include "pstree.h"
 #include "fsnotify.h"
 #include "kerndat.h"
+#include "vdso.h"
 
 #include "proc_parse.h"
 #include "protobuf.h"
@@ -96,9 +97,10 @@ static int parse_vmflags(char *buf, struct vma_area *vma_area)
 
 	do {
 		/* mmap() block */
-		if (_vmflag_match(tok, "gd"))
+		if (_vmflag_match(tok, "gd")) {
 			vma_area->vma.flags |= MAP_GROWSDOWN;
-		else if (_vmflag_match(tok, "lo"))
+			vma_area->vma.start -= PAGE_SIZE; /* Guard page */
+		} else if (_vmflag_match(tok, "lo"))
 			vma_area->vma.flags |= MAP_LOCKED;
 		else if (_vmflag_match(tok, "nr"))
 			vma_area->vma.flags |= MAP_NORESERVE;
@@ -166,11 +168,14 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 			goto err;
 	}
 
-	while (fgets(buf, BUF_SIZE, smaps)) {
+	while (1) {
 		int num;
 		char file_path[6];
+		bool eof;
 
-		if (!is_vma_range_fmt(buf)) {
+		eof = (fgets(buf, BUF_SIZE, smaps) == NULL);
+
+		if (!eof && !is_vma_range_fmt(buf)) {
 			if (!strncmp(buf, "Nonlinear", 9)) {
 				BUG_ON(!vma_area);
 				pr_err("Nonlinear mapping found %016"PRIx64"-%016"PRIx64"\n",
@@ -190,6 +195,21 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 				continue;
 		}
 
+		if (vma_area) {
+			list_add_tail(&vma_area->list, &vma_area_list->h);
+			vma_area_list->nr++;
+			if (privately_dump_vma(vma_area)) {
+				unsigned long pages;
+
+				pages = vma_area_len(vma_area) / PAGE_SIZE;
+				vma_area_list->priv_size += pages;
+				vma_area_list->longest = max(vma_area_list->longest, pages);
+			}
+		}
+
+		if (eof)
+			break;
+
 		vma_area = alloc_vma_area();
 		if (!vma_area)
 			goto err;
@@ -255,11 +275,13 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 		}
 
 		if (vma_area->vma.status != 0) {
-			goto done;
+			continue;
 		} else if (strstr(buf, "[vsyscall]")) {
 			vma_area->vma.status |= VMA_AREA_VSYSCALL;
 		} else if (strstr(buf, "[vdso]")) {
-			vma_area->vma.status |= VMA_AREA_REGULAR | VMA_AREA_VDSO;
+			vma_area->vma.status |= VMA_AREA_REGULAR;
+			if ((vma_area->vma.prot & VDSO_PROT) == VDSO_PROT)
+				vma_area->vma.status |= VMA_AREA_VDSO;
 		} else if (strstr(buf, "[heap]")) {
 			vma_area->vma.status |= VMA_AREA_REGULAR | VMA_AREA_HEAP;
 		} else {
@@ -317,16 +339,6 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 			}
 			vma_area->vma.flags  |= MAP_ANONYMOUS;
 		}
-done:
-		list_add_tail(&vma_area->list, &vma_area_list->h);
-		vma_area_list->nr++;
-		if (privately_dump_vma(vma_area)) {
-			unsigned long pages;
-
-			pages = vma_area_len(vma_area) / PAGE_SIZE;
-			vma_area_list->priv_size += pages;
-			vma_area_list->longest = max(vma_area_list->longest, pages);
-		}
 	}
 
 	vma_area = NULL;
@@ -1048,7 +1060,7 @@ int parse_fdinfo(int fd, int type,
 	pr_err("No records of type %d found in fdinfo file\n", type);
 parse_err:
 	ret = -1;
-	pr_perror("%s: error parsing [%s] for %d\n", __func__, str, type);
+	pr_perror("%s: error parsing [%s] for %d", __func__, str, type);
 out:
 	fclose(f);
 	return ret;
@@ -1130,6 +1142,7 @@ int parse_file_locks(void)
 			 */
 			pr_perror("We have a blocked file lock!");
 			ret = -1;
+			xfree(fl);
 			goto err;
 		}
 
@@ -1145,3 +1158,122 @@ err:
 	fclose(fl_locks);
 	return ret;
 }
+
+int parse_posix_timers(pid_t pid, struct proc_posix_timers_stat *args)
+{
+	int i;
+	int ret = 0;
+	int get = 0;
+	int pid_t;
+
+	FILE * file;
+	char * line1 = NULL;
+	char * line2 = NULL;
+	char * line3 = NULL;
+	char * line4 = NULL;
+	size_t len1 = 0;
+	size_t len2 = 0;
+	size_t len3 = 0;
+	size_t len4 = 0;
+
+	char * siginfo;
+	char siginfo_tmp[20];
+	char sigpid[7];
+	char tidpid[4];
+
+	char str_name[10];
+	struct proc_posix_timer *timer = NULL;
+
+	INIT_LIST_HEAD(&args->timers);
+	args->timer_n = 0;
+
+	file = fopen_proc(pid, "timers");
+	if (file == NULL) {
+		pr_perror("Can't open posix timers file!");
+		ret = -1;
+		goto end_posix;
+	}
+
+	while (1) {
+		get = getline(&line1, &len1, file);
+		if (get == -1)
+			goto end_posix;
+		get = getline(&line2, &len2, file);
+		if (get == -1)
+			goto end_posix;
+		get = getline(&line3, &len3, file);
+		if (get == -1)
+			goto end_posix;
+		get = getline(&line4, &len4, file);
+		if (get == -1)
+			goto end_posix;
+
+		timer = xzalloc(sizeof(struct proc_posix_timer));
+
+		ret = sscanf(line1, "%s %ld", str_name, &timer->spt.it_id);
+		if (ret != 2 || str_name[0] != 'I')
+			goto parse_err_posix;
+		ret = sscanf(line2, "%s %d%s", str_name, &timer->spt.si_signo, siginfo_tmp);
+		if (ret != 3 || str_name[0] != 's')
+			goto parse_err_posix;
+		siginfo=&siginfo_tmp[1];
+		ret = sscanf(siginfo, "%p", &timer->spt.sival_ptr);
+		if (ret != 1)
+			goto parse_err_posix;
+		for (i = 0; i<len3; i++) {
+			if (line3[i] == '/' || line3[i] == '.') {
+				line3[i] = ' ';
+			}
+		}
+		ret = sscanf(line3, "%s %s %s %d", str_name, sigpid, tidpid, &pid_t);
+		if (ret != 4 || str_name[0] != 'n')
+			goto parse_err_posix;
+
+		if ( tidpid[0] == 't') {
+			timer->spt.it_sigev_notify = SIGEV_THREAD_ID;
+		} else {
+			switch (sigpid[0]) {
+				case 's' :
+					timer->spt.it_sigev_notify = SIGEV_SIGNAL;
+					break;
+				case 't' :
+					timer->spt.it_sigev_notify = SIGEV_THREAD;
+					break;
+				default :
+					timer->spt.it_sigev_notify = SIGEV_NONE;
+					break;
+			}
+		}
+
+		ret = sscanf(line4, "%s %d", str_name, &timer->spt.clock_id);
+		if (ret != 2 || str_name[0] != 'C')
+			goto parse_err_posix;
+		list_add(&timer->list, &args->timers);
+		timer = NULL;
+		args->timer_n++;
+	}
+parse_err_posix:
+	while (!list_empty(&args->timers)) {
+		timer = list_first_entry(&args->timers, struct proc_posix_timer, list);
+		list_del(&timer->list);
+		xfree(timer);
+	}
+	pr_perror("Parse error in posix timers proc file!");
+	ret = -1;
+end_posix:
+	if (ferror(file)) {
+		ret = -1;
+		pr_perror("getline");
+	}
+
+	if (line1)
+		free(line1);
+	if (line2)
+		free(line2);
+	if (line3)
+		free(line3);
+
+	if (file != NULL)
+		fclose(file);
+	return ret;
+}
