@@ -42,7 +42,7 @@ struct packet_sock_desc {
 
 #define NO_FANOUT	((unsigned int)-1)
 
-void show_packetsk(int fd, struct cr_options *o)
+void show_packetsk(int fd)
 {
 	pb_show_plain_pretty(fd, PB_PACKETSK, "5:%d");
 }
@@ -205,16 +205,11 @@ out:
 	return ret;
 }
 
-static const struct fdtype_ops packet_dump_ops = {
+const struct fdtype_ops packet_dump_ops = {
 	.type		= FD_TYPES__PACKETSK,
 	.dump		= dump_one_packet_fd,
 };
 
-int dump_one_packet_sk(struct fd_parms *p, int lfd, const int fdinfo)
-{
-	return do_dump_gen_file(p, lfd, &packet_dump_ops, fdinfo);
-}
-
 int dump_socket_map(struct vma_area *vma)
 {
 	struct packet_sock_desc *sd;
