@@ -251,16 +251,11 @@ err:
 	return -1;
 }
 
-static const struct fdtype_ops unix_dump_ops = {
+const struct fdtype_ops unix_dump_ops = {
 	.type		= FD_TYPES__UNIXSK,
 	.dump		= dump_one_unix_fd,
 };
 
-int dump_one_unix(struct fd_parms *p, int lfd, const int fdinfo)
-{
-	return do_dump_gen_file(p, lfd, &unix_dump_ops, fdinfo);
-}
-
 static int unix_collect_one(const struct unix_diag_msg *m,
 		struct rtattr **tb)
 {
@@ -491,7 +486,7 @@ static struct unix_sk_info *find_unix_sk_by_ino(int ino)
 	return NULL;
 }
 
-void show_unixsk(int fd, struct cr_options *o)
+void show_unixsk(int fd)
 {
 	pb_show_plain_pretty(fd, PB_UNIXSK, "1:%#x 2:%#x 3:%d 4:%d 5:%d 6:%d 7:%d 8:%#x 11:S");
 }
