@@ -80,6 +80,7 @@ struct tty_info {
 	int				major;
 
 	bool				create;
+	bool				inherit;
 };
 
 struct tty_dump_info {
@@ -95,8 +96,6 @@ struct tty_dump_info {
 static LIST_HEAD(all_tty_info_entries);
 static LIST_HEAD(all_ttys);
 
-#define INHERIT_SID			(-1)
-
 /*
  * Usually an application has not that many ttys opened.
  * If this won't be enough in future we simply need to
@@ -200,7 +199,7 @@ static int parse_index(u32 id, int lfd, int major)
 	switch (major) {
 	case TTYAUX_MAJOR:
 		if (ioctl(lfd, TIOCGPTN, &index)) {
-			pr_perror("Can't obtain ptmx index\n");
+			pr_perror("Can't obtain ptmx index");
 			return -1;
 		}
 		break;
@@ -328,7 +327,7 @@ static int lock_pty(int fd)
 static int tty_set_sid(int fd)
 {
 	if (ioctl(fd, TIOCSCTTY, 1)) {
-		pr_perror("Can't set sid on terminal fd %d\n", fd);
+		pr_perror("Can't set sid on terminal fd %d", fd);
 		return -1;
 	}
 
@@ -338,7 +337,7 @@ static int tty_set_sid(int fd)
 static int tty_set_prgp(int fd, int group)
 {
 	if (ioctl(fd, TIOCSPGRP, &group)) {
-		pr_perror("Failed to set group %d on %d\n", group, fd);
+		pr_perror("Failed to set group %d on %d", group, fd);
 		return -1;
 	}
 	return 0;
@@ -411,9 +410,9 @@ static bool tty_has_active_pair(struct tty_info *info)
 
 static void tty_show_pty_info(char *prefix, struct tty_info *info)
 {
-	pr_info("%s type %s id %#x index %d (master %d sid %d pgrp %d)\n",
+	pr_info("%s type %s id %#x index %d (master %d sid %d pgrp %d inherit %d)\n",
 		prefix, tty_type(info->major), info->tfe->id, info->tie->pty->index,
-		pty_is_master(info), info->tie->sid, info->tie->pgrp);
+		pty_is_master(info), info->tie->sid, info->tie->pgrp, info->inherit);
 }
 
 static int restore_tty_params(int fd, struct tty_info *info)
@@ -535,7 +534,7 @@ static int pty_open_unpaired_slave(struct file_desc *d, struct tty_info *slave)
 	 * be inherited, either it requires a fake master.
 	 */
 
-	if (likely(slave->tie->sid == INHERIT_SID)) {
+	if (likely(slave->inherit)) {
 		fd = dup(get_service_fd(SELF_STDIN_OFF));
 		if (fd < 0) {
 			pr_perror("Can't dup SELF_STDIN_OFF");
@@ -577,9 +576,20 @@ static int pty_open_unpaired_slave(struct file_desc *d, struct tty_info *slave)
 	 * be already restored properly thus we can simply
 	 * use syscalls instead of lookup via process tree.
 	 */
-	if (likely(slave->tie->sid == INHERIT_SID)) {
-		if (tty_set_prgp(fd, getpgid(getppid())))
-			goto err;
+	if (likely(slave->inherit)) {
+		/*
+		 * The restoration procedure only works if we're
+		 * migrating not a session leader, otherwise it's
+		 * not allowed to restore a group and one better to
+		 * checkpoint complete process tree together with
+		 * the process which keeps the master peer.
+		 */
+		if (root_item->sid != root_item->pid.virt) {
+			pr_debug("Restore inherited group %d\n",
+				 getpgid(getppid()));
+			if (tty_set_prgp(fd, getpgid(getppid())))
+				goto err;
+		}
 	}
 
 	if (pty_open_slaves(slave))
@@ -763,7 +773,7 @@ static int tty_find_restoring_task(struct tty_info *info)
 shell_job:
 	if (opts.shell_job) {
 		pr_info("Inherit terminal for id %x\n", info->tfe->id);
-		info->tie->sid = info->tie->pgrp = INHERIT_SID;
+		info->inherit = true;
 		return 0;
 	}
 
@@ -944,6 +954,7 @@ static int collect_one_tty(void *obj, ProtobufCMessage *msg)
 	INIT_LIST_HEAD(&info->sibling);
 	info->major = major(info->tie->rdev);
 	info->create = (info->major == TTYAUX_MAJOR);
+	info->inherit = false;
 
 	if (verify_info(info))
 		return -1;
@@ -1185,19 +1196,14 @@ static int dump_one_pty(int lfd, u32 id, const struct fd_parms *p)
 	return ret;
 }
 
-static const struct fdtype_ops tty_ops = {
+const struct fdtype_ops tty_dump_ops = {
 	.type	= FD_TYPES__TTY,
 	.dump	= dump_one_pty,
 };
 
-int dump_tty(struct fd_parms *p, int lfd, const int fdinfo)
+int tty_prep_fds(void)
 {
-	return do_dump_gen_file(p, lfd, &tty_ops, fdinfo);
-}
-
-int tty_prep_fds(struct cr_options *opts)
-{
-	if (!opts->shell_job)
+	if (!opts.shell_job)
 		return 0;
 
 	if (!isatty(STDIN_FILENO)) {
